require("common")

-- The main settings skeleton
settings = {
    base = {
        -- Base platform settings, that copied to work set in load_from_storage() function
        -- Settings without values (valur = nil) will be removed from replies
        { name = "general/online_mode", meta = {type = "bool"} },
        { name = "night_vision/mode", meta = {type = "choice", range = {"on", "off", "auto"}} },
        { name = "microphone/enabled", meta = {type = "bool"} },
        { name = "microphone/sensitivity", meta = {type = "int", range = {0, 20}} },
        { name = "led/enabled", meta = {type = "bool"} },
        { name = "general/power_freq", meta = {type = "choice", range = {"50hz", "60hz", "auto"}} },
        { name = "sound_detection/enabled", meta = {type = "bool"} },
        { name = "sound_detection/sensitivity", meta = {type = "int", range = {0, 100}} },
        { name = "motion_detector/enabled", meta = {type = "bool"} },
        { name = "general/timezone", meta = {type = "choice"} },

        { name = "image/brightness", meta = {type = "int", range = {0, 100} } },
        { name = "image/contrast", meta = {type = "int", range = {0, 100} } },
        { name = "image/saturation", meta = {type = "int", range = {0, 100} } },
        { name = "image/hue", meta = {type = "int", range = {0, 100} } },
        { name = "image/upside_down", meta = {type = "bool"} },
        { name = "image/rotate", meta = {type = "bool"} },
        { name = "image/wb", meta = {type = "choice"} }, -- manual can be adjusted by wb_red + wb_blue exact setting
        { name = "image/shutter_time", meta = {type = "choice"} },
        { name = "image/gain", meta = {type = "int", range = {}} },
        { name = "image/wdr", meta = {type = "bool"} }
    },
    tmp = {
        -- All camera settings. If some cloud restrictions are defined for the specific setting then it may
        -- has the 'value' and 'meta' fields. Anyway, you can modify them all
    },
    traits = {
        -- Place here all needed traits for settings, Example:
        -- ["streams/0/resolution"] = {
        --     depends = {"streams/0/fps", "streams/0/bitrate"},    -- Depended settings array
        --     restart_requered = true,     -- Is restart required after setting appliing
        --     get = getter_func,           -- Getter function to get actual setting value from cam
        --     set = setter_func,           -- Setter function to set new value to cam
        -- Optional fields:
        --     is_changed_hint = changed_func   -- Function, which takes the current value of the setting
                                                -- and returns 'true' if the setting value is really changed
                                                -- and 'false' otherwise
        -- },

        access_funcs_owner_table = nil      -- if getteres and setters are placed in some table, you must init this field. It's value will be passed as an single argument to setters and getters functions
    }
}

-- Add new settings into default array
function settings:add(new_settings)
    for _, v in pairs(new_settings) do
        for i, s in ipairs(self.base) do if s.name == v.name then table.remove(self.base, i) ; break end end -- replace case
        self.base[#self.base+1] = v
    end
end

for i = 0, 2 do -- triple streaming!
    local prefix = "streams/".. tostring(i)
    settings:add({
    { name = prefix .. "/resolution", meta = {type = "choice"} },
    { name = prefix .. "/bitrate", meta = {type = "choice"} },
    { name = prefix .. "/fps", meta = {type = "choice"} },
    { name = prefix .. "/i_frame_interval", meta = {type = "int"} },
    { name = prefix .. "/variable_bitrate", meta = {type = "bool"} }, -- CBR / VBR
    { name = prefix .. "/codec", meta = {type = "choice"} },
    { name = prefix .. "/acodec", meta = {type = "choice"} }
    })
end


-- Get traits for setting with name 'setting_name' or creates it
function settings:__call(setting_name)
    if not self.traits[setting_name] then self.traits[setting_name] = {} end
    return self.traits[setting_name]
end

-- Get setting by name
function settings:__index(setting_name)
    for _, s in ipairs(self.tmp) do
        if s.name == setting_name then return s end
    end

    return nil
end

function settings:remove_base(setting_name)
    self.traits[setting_name] = nil
    for i, s in ipairs(self.base) do
        if s.name == setting_name then return table.remove(self.base, i) end
    end
end

-- Remove setting with name 'setting_name'
-- Note: It is NOT SAFE to call this function inside the loop
-- over self.base and self.shadow tables
function settings:remove(setting_name)
    for i, s in ipairs(self.tmp) do
        if s.name == setting_name then
            table.remove(self.tmp, i) -- remove only from tempotary table!
            s.removed = true
            self:remove_from_cache(setting_name)
            info_log(setting_name .. " was removed")
            return s
        end
    end
end

function settings:clear_values()
    self.tmp = table.copy(self.base)
end

local function on_handler_err(msg)
    err_log(debug.traceback(msg, 2))
end

-- Base implementation of load_all
function settings:update(update_config)
    local update_set = update_config
    if update_set then
        info_log("Updating settings: " .. json.write(update_set))
    else
        info_log("Updating all settings... ")
        self:clear_values()
        update_set = self.tmp
    end

    local invoked_getters = {}
    local i = 1
    while i <= #update_set do
        -- Perform the safe traverse over the settings, because it is possible that some of them will be removed from the getters
        local s = update_set[i]
        local getter = self(s.name).get
        if getter and not invoked_getters[getter] then
            info_log("--> loading '" .. s.name .. "'")
            xpcall(getter, on_handler_err, self.traits.access_funcs_owner_table)
            invoked_getters[getter] = true
            collectgarbage()
        end
        if s.removed then
            assert(s ~= update_set[i]) -- self.tmp must be already shifted
        else
            i = i + 1
        end
    end
end

function settings:remove_from_cache(setting_name)
    local camera = nil  -- todo: supply a valid camera id or remove this function
    if camera == nil then return end

    local config = cache_storage:get_cmd_data_by_cam_id("config", camera)
    if type(config) ~= "table" then
        return
    end

    local _, i = find_setting(config, setting_name)
    if not i then
        return
    end

    table.remove(config, i)
    cache_storage:set_cmd_data_by_cam_id("config", camera, config)
end

function settings:load_from_storage(camera)
    local new_config = cache_storage:get_cmd_data_by_cam_id("config", camera)

    if type(new_config) == "table" and #new_config ~= 0 then
        self.tmp = new_config
    else
        self.tmp = table.copy(self.base)
    end

    return self.tmp
end

function settings:save_to_storage(camera)
    -- NB: If we need to up-to-date cloud config - we need to force config.update command
    return cache_storage:set_cmd_data_by_cam_id("config", camera, self.tmp) -- true/false
end

setmetatable(settings, settings)

-- Cloud config utilities

-- Safe call of method (exported function)
function safe_call(func, ...)
    return xpcall(func, on_handler_err, ...)
end

function find_setting(settings_array, setting_name)
    for i, s in ipairs(settings_array) do
        if s.name == setting_name then
            return s, i
        end
    end
    return nil
end

local function add_dependent_settings(response_settings)
    for _, resp_setting in ipairs(response_settings) do
        local dep = settings(resp_setting.name).depends
        if dep ~= nil then
            for _, dep_setting_name in ipairs(dep) do
                if not find_setting(response_settings, dep_setting_name) then
                    response_settings[#response_settings+1] = {
                        name = dep_setting_name, type = resp_setting.type, camera = resp_setting.camera
                    }
                end
            end
        end
    end
end

local function false_func()
    return false
end

local function is_value_changed(setting, setting_curr_val)
    -- Here we get a custom checker (if it exists) from the implementation and compare the
    -- setting's value with the current (i.e. with the cached one) and call the custom checker
    local is_changed = settings(setting.name).is_changed_hint or false_func
    if setting.value ~= setting_curr_val or is_changed(setting_curr_val) then
        return true
    end
    return false
end

local function parse_config_set(config, handlers_to_run, response_settings)
    for _, upd_setting in ipairs(config) do
        local setting = settings[upd_setting.name]
        if setting == nil then
            warn_log("Skipping setting '"..upd_setting.name.."': not found in template")
            goto next_setting
        end

        -- This table will be sent in config.update.reply. Event if this setting has a bad value
        -- we include it to the response and later put it's actual value into the reply
        response_settings[#response_settings+1] = upd_setting

        if upd_setting.value == nil then
            warn_log("Skipping setting '"..upd_setting.name.."': undefined value")
            goto next_setting
        end

        -- 'setting' table holds the actual value, which is taken from the cache
        if not is_value_changed(upd_setting, setting.value) then
            -- The setting stays unchanged
            info_log("Skipping setting '"..upd_setting.name.."': unchanged value")
            goto next_setting
        end

        -- Check if the setting suits for its range
        if setting.meta.range ~= nil and setting.meta.type == "choice" then -- int has special 'range' field
            if not is_in_table(upd_setting.value, setting.meta.range) then
                err_log(string.format("Skipping setting '%s': value '%s' is out of range: %s ",
                    upd_setting.name, tostring(upd_setting.value), json.write(setting.meta.range)
                ))
                goto next_setting
            end
        end

        -- Save new value in the settings table and add special handler function that applies this value to the device
        setting.value = upd_setting.value
        setting.changed = true -- it will be flushed in clear_values()

        local current_handler = settings(upd_setting.name).set;
        if current_handler ~= nil then
            if handlers_to_run[current_handler] == nil then handlers_to_run[current_handler] = {} end
            handlers_to_run[current_handler][ #(handlers_to_run[current_handler])+1 ] = upd_setting.name
        end
        ::next_setting::
    end

    -- Put to the response_settings all the dependent settings, which have not been included yet
    add_dependent_settings(response_settings)
end

local function create_reply_config(config)
    local result = json.array()
    for _, setting in ipairs(config) do
        -- Exclude undefined values from the reply config
        if setting.value ~= nil then result[#result + 1] = setting end
    end
    return result
end

-- Cloud config get handler
function cloud_config_get(args)
    if args.camera then
        if type(setup.set_camera) == "function" then
            setup:set_camera(args.camera)
        else
            settings:load_from_storage(args.camera)
        end
    end
    settings:update()
    return create_reply_config(settings.tmp)
end

local function add_other_changed_settings(settings_cache, update_config, response_settings)
    -- Compare just loaded fresh settings with the cached ones earlier.
    -- If we found a setting, which is not present in cache (i.e. a new setting)
    -- or it's value is different with the cached one (i.e. a new setting value)
    -- and that setting is not present in update_config, then we'll include
    -- it to the update_config.
    -- The logic is correct even if the cache has already been staled or doesn't exist
    -- TODO: check if meta changes

    if type(settings_cache) ~= "table" then return end

    local changed_settings = {}
    for _, setting in ipairs(settings.tmp) do
        local cached_setting_value = nil
        if settings_cache[setting.name] ~= nil then
            cached_setting_value = settings_cache[setting.name].value
        end

        if setting.value ~= nil and not find_setting(update_config, setting.name) and not find_setting(response_settings, setting.name) then
            if type(setting.value) ~= "table" and setting.value ~= cached_setting_value then
                changed_settings[setting] = true
            end
            if type(setting.value) == "table" then
                for k, v in pairs(setting.value) do if type(cached_setting_value) ~= "table" or type(cached_setting_value[k]) ~= type(v) or cached_setting_value[k] ~= v then
                    changed_settings[setting] = true
                    break
                end end
            end
        end
    end

    for s, _ in pairs(changed_settings) do
        response_settings[#response_settings + 1] = s
    end
end

-- Cloud config update handler
function cloud_config_update(args)
    local config = args.config
    if type(config) == "string" then config = json.parseStr(config) end
    if #config == 0 then return json.array() end

    local used_handlers = {} -- handlers in format: ["<handler func>": [setting1, setting2, setting3]]
    local response_settings = {} -- response table
    local settings_cache = {}
    local camera_id = config[1].camera
    if caps.requests.config.useCache then
        if type(setup.set_camera) == "function" then setup:set_camera(camera_id) else settings:load_from_storage(camera_id) end
        settings_cache = table.copy(settings)
    end

    parse_config_set(config, used_handlers, response_settings)

    -- Apply new settings and remove settings that can't be set
    for handler_func, settings_in_handler in pairs(used_handlers) do
        local ok, h_success = xpcall(handler_func, on_handler_err, settings.traits.access_funcs_owner_table)
        if ok and h_success then
            -- Handler processed succesfully and new settings are set on camera:
            for i = 1, #settings_in_handler do
                if settings(settings_in_handler[i]).restart_requered then
                    local setting = find_setting(response_settings, settings_in_handler[i])
                    setting.is_stream_restarted = true
                end
            end
        else
            err_log("Failed to apply handler for next settings (they will be removed from reply): " .. json.write(settings_in_handler))
            for _, setting_name in ipairs(settings_in_handler) do
                local _, setting_index = find_setting(response_settings, setting_name)
                table.remove(response_settings, setting_index)
            end
        end
        collectgarbage()
    end

    if caps.requests.config.updateAllOnSync then
        settings:update()
    else
        settings:update(config)
    end

    for _, resp_setting in ipairs(response_settings) do
        if not settings[resp_setting.name] then
            resp_setting = nil
        else
            resp_setting.value = settings[resp_setting.name].value
            resp_setting.meta = settings[resp_setting.name].meta
            resp_setting.camera = camera_id
        end
    end

    if caps.requests.config.useCache then
        -- If non-requested settings have been changed, we send it to the cloud
        add_other_changed_settings(settings_cache, config, response_settings)
    end

    return create_reply_config(response_settings)
end

function cloud_config(cmd)
    if cmd["cmd"] == "config.get" then
        return cloud_config_get(cmd.args)
    elseif cmd["cmd"] == "config.update" then
        return cloud_config_update(cmd.args)
    end
end

-- Requests traits
caps = {
    requests = {
        config = {
            handler = "cloud_config",
            useCache = true,
            updateAllOnSync = true
        }
    }
}

function init()
    log_config.tag = tostring(init_data.streamIndex or "-") -- init_data.name for DVR / IP camera name from config
    if type(setup_api_request_params) == "function" then setup_api_request_params(init_data.cdiUrl) end

    if type(init_data.channels) ~= "table" then
        iv.log("init: empty global init_data!")
        return
    end
    for k,v in pairs(init_data.channels) do
        -- ex: k = "0", v = {localId="0",streamIndex=327680,url="rtsp://login:pwd@ip/path"}
    end
    if type(setup_settings_handlers) == "function" then setup_settings_handlers() end

    if init_data.vendor and #(init_data.vendor) ~= 0 then settings.vendor_name = init_data.vendor end
    if init_spec then init_spec() end -- Perform project-specific initialization

    -- At the end of initialization perform setup of stop function
    if init_data.stop_func then http_request.cancel_func = init_data.stop_func end
end

md = {
    raw_width, raw_height, real_width = nil, -- !NB initialize it in init() func
    hex_map = {
        ["0"] = "0000", ["1"] = "0001", ["2"] = "0010", ["3"] = "0011", ["4"] = "0100", ["5"] = "0101", ["6"] = "0110", ["7"] = "0111",
        ["8"] = "1000", ["9"] = "1001", ["A"] = "1010", ["B"] = "1011", ["C"] = "1100", ["D"] = "1101", ["E"] = "1110", ["F"] = "1111"
    },
    bin_map = nil,
    ivideon_height = 8,
    ivideon_width = 12,
    width_k = 1,
    height_k = 1
}

function md:hex2bin(hexstr, trunc)
    local bin = hexstr:upper():gsub("[0-9A-F]", self.hex_map)
    if trunc then
        while bin:sub(1,1) == "0" do bin = bin:sub(2) end
    end
    return bin
end
function md:bin2hex(binstr)
    if not self.bin_map then self.bin_map = {} ; for k, v in pairs(self.hex_map) do self.bin_map[v] = k end end

    local result = ""
    for i = 0, #binstr - 1 do if i % 4 == 0 then
        if i + 4 > #binstr then result = self.bin_map[ string.rep("0", i + 4 - #binstr) .. binstr:sub(1, #binstr - i) ] .. result
        else result = self.bin_map[ binstr:sub(#binstr - i - 3, #binstr - i) ] .. result end
    end end
    return result
end

function string.tohex(binstr)
    return (binstr:gsub('.', function (c) return string.format('%02X', string.byte(c)) end))
end

function md:extract_window_from_bingrid(bingrid)
    local function strsym(str, char_number) return str:sub(char_number + 1, char_number + 1) end -- 0 char is 1st char in Lua
    local begin_point = {x = 0, y = 0, width = 0, height = 0}
    local end_point = {x = 0, y = 0, width = 0, height = 0}

    local area_found = false
    for y = 0, md.raw_height-1 do for x = 0, md.raw_width-1 do
        if not area_found then if strsym(bingrid, y * md.raw_width + x) == "1" then begin_point.x = x; begin_point.y = y; end_point.x = x; area_found = true end
        else if strsym(bingrid, y * md.raw_width + x) == "1" then end_point.x = x else break end end
    end if area_found then break end end
    if not area_found then return 0, 0, 0, 0 end

    area_found = false
    for y = begin_point.y + 1, md.raw_height-1 do for x = begin_point.x, end_point.x do
        if strsym(bingrid, y * md.raw_width + x) == "0" then end_point.y = y - 1; area_found = true; break end
    end if area_found then break end end
    if not area_found then end_point.y = md.raw_height-1 end

    return begin_point.x,
           begin_point.y,
           end_point.x - begin_point.x + 1,
           end_point.y - begin_point.y + 1
end

function md:win_to_bingrid(win_x, win_y, width, height)
    local result = ""
    for y = 0, md.raw_height-1 do for x = 0, md.raw_width-1 do
        if (y >= win_y and y < win_y + height) and x >= win_x and x < win_x + width then
            result = result .. "1"
        else
            result = result .. "0"
        end
    end end
    return result
end

function md:get_iv_legacy_window() -- return x, y, width, height
    local x, y, width, height = md:extract_window_from_bingrid(settings["motion_detector/0/zones"].value or "")
    local result = {
        round(x / md.width_k),
        round(y / md.height_k),
        round(width / md.width_k),
        round(height / md.height_k)
    }

    if result[1] >= md.ivideon_width then
        result[1] = md.ivideon_width - 1   -- out of band window x coord
    end

    if result[2] >= md.ivideon_height then
        result[2] = md.ivideon_height - 1  -- out of band window y coord
    end

    if result[1] + result[3] > md.ivideon_width then
        result[3] = md.ivideon_width - result[1] -- out of band window width
    elseif result[3] < 1 then
        result[3] = 1   -- too small window width
    end

    if result[2] + result[4] > md.ivideon_height then
        result[4] = md.ivideon_height - result[2]  -- out of band window height
    elseif result[4] < 1 then
        result[4] = 1 -- too small window height
    end

    return result
end
