local function hide_password(url)
    -- Searching for match for lua pattern with 3 captures: "(scheme)://(user):(password)@"
    return url:gsub("(.+)://(.+):(.+)@", "%1://%2:***@", 1)
end

-- A http_request callable table with call params holding (as memfunction)
http_request = {
    timeout_ms = 2000,
    max_retries_count = 5,
    retry_delay_ms = 500,
    base_url = "",
    cancel_func = function() return false end,

    -- HTTP errors
    OK_CODE = 200,
    BAD_REQUEST_CODE = 400
}

function http_request:__call(anyUrl, method, requestData, timeout)
    local url = ""
    if anyUrl:find("http://") or anyUrl:find("https://") then
        url = anyUrl
    else -- relative url
        if string.is_empty(self.base_url) then error("There are no baseUrl for req " .. anyUrl) end

        if self.base_url:sub(-1) == "/" and anyUrl:sub(1,1) == "/" then
            anyUrl = anyUrl:sub(2)
        elseif self.base_url:sub(-1) ~= "/" and anyUrl:sub(1,1) ~= "/" then
            anyUrl = "/" .. anyUrl
        end
        url = self.base_url .. anyUrl
    end

    local status, response
    for i=1, self.max_retries_count do
        if self.cancel_func() then
            return nil, self.BAD_REQUEST_CODE
        end

        if method == nil then
            status, response = http.get{url=url, timeout=timeout or self.timeout_ms}
        else
            status, response = http.performRequest{url=url, method=method, input=requestData, timeout=timeout or self.timeout_ms}
        end

        if status and ( response.code == self.OK_CODE or (#response.output > 0 and not response.output:find("Error")) ) then
            return response.output, (response.code or self.OK_CODE)
        end

        if self.cancel_func() then
            return nil, self.BAD_REQUEST_CODE
        end

        if i ~= self.max_retries_count then
            iv.sleep(self.retry_delay_ms)
        end
    end

    local verbose_error_msg=""
    if response then
        verbose_error_msg = "Code: "..tostring(response.code).." ("..response.codeDescr..")."
        if response.output:len() ~= 0 then
            verbose_error_msg = verbose_error_msg.." Output: "..response.output
        end
    end

    iv.log("Request to " .. hide_password(url) .. " failed. "..verbose_error_msg, {severity = "ERROR"})
    return nil, (response and response.code or self.BAD_REQUEST_CODE)
end

function http_request:set_base_url(httpUrl)
    local parsed_url = iv.parse_url(httpUrl)
    self.base_url = iv.build_url {
        scheme = parsed_url.scheme,
        userinfo = parsed_url.userinfo or "",
        host = parsed_url.host,
        port = parsed_url.port
    }
    self.base_url = self.base_url:sub(1,-2) -- trunc endslash
end

function http_request:copy_url_with(params)
    return iv.copy_url_with(self.base_url, params)
end

setmetatable(http_request, http_request)

function iv.copy_url_with(url, params)
    local parsed_url = iv.parse_url(url)
    for k,v in pairs(params) do
        parsed_url[k] = v
    end

    return iv.build_url(parsed_url)
end

function popen(program, mode)
    local success, handler = pcall(io.popen, program, mode)
    if success then
        return handler
    else
        return nil
    end
end

function string.is_empty(str)
    return str == nil or str == ""
end

function string:split(sep)
    local sep, fields = sep or ":", {}
    local pattern = string.format("([^%s]+)", sep)
    self:gsub(pattern, function(c) fields[#fields+1] = c end)
    return fields
end

math.randomseed( os.time() )
function string:at(i) return self:sub(i, i) end -- works only for ANSII
local print_chars = "qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890"
function string.random(len, chars)
    if not chars then chars = print_chars end
    local rand_char = chars:at( math.random(1, #chars) )
    if len > 1 then return string.random(len - 1) .. rand_char end
    return rand_char
end

function is_in_table(value, t) -- bool
    if #t == 0 then -- it's a hash table
        if t[value] ~= nil then return true end
    else -- it's just an array
        for i = 1, #t do
            if value == t[i] then return true end
        end
    end

    return false
end

function is_file_exist(path)
    local f = io.open(path, "r")
    if f ~= nil then
        f:close()
        return true
    end
    return false
end

log_config = {
    tag = nil
}

local function log_tag()
    if log_config.tag then
        return "Lua ["..log_config.tag.."]"
    end

    return "Lua"
end

function err_log(msg)
    iv.log(msg, {severity = "ERROR", tag = log_tag()})
end

function warn_log(msg)
    iv.log(msg, {severity = "WARNING", tag = log_tag()})
end

function info_log(msg)
    iv.log(msg, {severity = "INFO", tag = log_tag()})
end

-- Return an iterator function for the difference between t1 and t2: every call of
-- such iterator function returns the next element from t1, which is not present in t2
function arrays_diff(t1, t2, cmp_func)
    if not cmp_func then
        cmp_func = function(v1, v2) return v1 == v2 end
    end

    local iterator = function(unused, index)
        for i = index+1, #t1 do
            local found = false
            for _, v in ipairs(t2) do
                if cmp_func(t1[i], v) then
                    found = true
                end
            end

            if not found then
                return i, t1[i]
            end
        end
        return nil
    end
    return iterator, nil, 0
end

function dicts_diff(t1, t2, cmp_func)
    if not cmp_func then
        cmp_func = function(v1, v2) return true end
    end

    local iterator = function(unused, index)
        while true do
            local k, v = next(t1, index)
            if not k then return nil end
            index = k
            local t2_val = t2[k]
            if not t2_val or not cmp_func(v, t2_val) then
                return k, v
            end
        end
    end
    return iterator, nil, nil
end

function r_ipairs(t)
    local index = #t + 1
    return function(t)
        index = index - 1
        if index < 1 or index > #t then
            return nil
        end
        return index, t[index]
    end, t, nil
end

function table.is_subtable_seq(src_table, ...)
    local tt = src_table
    for _, v in pairs({...}) do
        if type(tt) ~= "table" then return false end
        tt = tt[v]
    end
    return (type(tt) == "table")
end

function table.merge(dst, src, append_lists)
    local function is_list(t) -- numeric-indexed table (w/o keys)
        for k, _ in pairs(t) do if tonumber(k) == nil then return false end end
        return true
    end

    if type(dst) ~= "table" or type(src) ~= "table" then return end

    -- Disable lists appending if we haven't lists:
    local do_append = append_lists and is_list(src) and is_list(dst)

    for k, v in pairs(src) do
        if do_append then
            table.insert(dst, v)
        else
            if type(v) == "table" and type(dst[k] or false) == "table" then
                table.merge(dst[k], v, append_lists)
            else
                dst[k] = v
            end
        end
    end
end

function table.getn(src) -- return total (index + keys) length of dict
    if type(src) ~= "table" then return nil end
    local count = 0
    for k, _ in pairs(src) do count = count + 1 end
    return count
end

function table.copy(src, rl) -- rl - recursion limit, default nil
    rl = rl or 10
    local orig_type = type(src)
    local copy = {}
    if orig_type == 'table' then
        for k, v in pairs(src) do copy[table.copy(k, rl)] = table.copy(v, rl) end
        setmetatable( copy, getmetatable(src) )
    else -- number, string, boolean, etc
        copy = src
    end
    return copy
end

-- Time funcs
SEC_IN_DAY = 86400
SEC_IN_HOUR = 3600
SEC_IN_MIN = 60

MS_IN_DAY = 86400000
MS_IN_HOUR = 3600000
MS_IN_MIN = 60000
MS_IN_SEC = 1000

function to_period_begin(ts, period_duration)
    return ts - ts % period_duration
end

function to_period_end(ts, period_duration)
    return to_period_begin(ts, period_duration) + period_duration - 1
end

function to_next_period(ts, period_duration)
    return to_period_begin(ts, period_duration) + period_duration
end

--math funcs

-- Math round function with round degree
-- round(0.4551) --> 0  round to integer
-- round(4.4551) --> 4  round to integer
-- round(1455.4551, -2) --> 1500.0  2 digital BEFORE point
-- round(0.4551, 2) --> 0.46  2 digit AFTER point
function round(num, numDecimalPlaces)
    local mult = 10^(numDecimalPlaces or 0)
    local result = nil
    if num >= 0 then result = math.floor(num * mult + 0.5)
    else result = math.ceil(num * mult - 0.5) end

    if numDecimalPlaces ~= nil then result = result / mult end -- division always makes float values!
    return result
end

-- function round(n) return math.floor( n + 0.5 ) end