#!/bin/bash

g_services=("ivideon-server" "remote-svc")
g_service=
server_config_subpath=".IvideonServer"
systemd_path="/etc/systemd/system"

# Formatting:
bold_format=$(tput bold)
normal_format=$(tput sgr0)

usage() {
    local message="${1-}"
    [ "${message}" ] && echo -e "Error: ${bold_format}${message}${normal_format}\n"
    echo "Usage: $0 [COMMAND] [SERVICE] [<user> [<config> [<dump_dir> [<sv_suffix>]]]]"
    echo
    echo "Supported COMMANDs are:"
    echo "is-service Check is the service registered in system"
    echo "register   Just install system service without autorun"
    echo "           ${bold_format}<user>${normal_format} argument is required"
    echo "check      Check is the service configured for autorun"
    echo "enable     To turn autostart on for already registered service"
    echo "install    Install with system autorun"
    echo "           ${bold_format}<user>${normal_format} argument is required"
    echo "install_and_run The same action with instant start of the service"
    echo "disable    Turns off autostart of service(s). Can take few services at once."
    echo "           If services aren't specified, ${bold_format}all supported services${normal_format} will be disabled"
    echo "stop_and_disable The same action with safe stop before disabling"
    echo "uninstall  Stop and remove from system. Can take few services at once."
    echo "           If services aren't specified, ${bold_format}all supported services${normal_format} will be uninstalled"
    echo 
    echo "Supported SERVICEs: ${bold_format}${g_services[@]}${normal_format}"
    echo
    echo "Possible ARGUMENTs:"
    echo "user      Mandatory for setup commands. Linux username for all operations. Must have complete"
    echo "          networks and devices access that you're wants to use"
    echo "config    Absolute path to config file. \"~/${server_config_subpath}/videoserverd.config\" by default"
    echo "dump_dir  Absolute path to crash-dumps folder. \"~/${server_config_subpath}/dumps\" by default"
    echo "sv_suffix NOT RECOMMENDED service name suffix to create multiple services for different users or config files."
    echo "          Will be added to the end of service name. Will not be managed automatically using GUI or further updated"
    exit 1
}

ctlcmd() {
    systemctl --no-pager ${1-} ${g_service}.service
}

is_service_supported() {
    local s="${1}"
    echo "${g_services[@]}" | grep --quiet --word-regexp "${s}" || usage "Service \"${s}\" unsupported"
}

check_service() {
    ctlcmd cat > /dev/null
}

check_autorun() {
    ctlcmd is-enabled
}

add_sudoers() {
    local username=${1-}

    echo -e "\n\
${username} ALL=(ALL:ALL) NOPASSWD: /bin/systemctl start ivideon-server.service, /bin/systemctl stop ivideon-server.service\
" > /tmp/${username}-ivideon.conf
    chmod 0440 /tmp/${username}-ivideon.conf
    mv /tmp/${username}-ivideon.conf /etc/sudoers.d/${username}-ivideon
}

do_register() {
    local username="${1}"
    local absolute_path_to_config="${2-}"
    local absolute_path_to_dumps="${3-}"
    local service_suffix="${4-}"

    [ -z "${username}" ] && usage "Please, provide username and path to videoserver config"

    local homedir="$(getent passwd "${username}" | cut -d: -f6)"

    if [ -z "${absolute_path_to_config}" ]; then
        absolute_path_to_config="${homedir}/${server_config_subpath}/videoserverd.config"
    fi

    if [ -z "${absolute_path_to_dumps}" ]; then
        absolute_path_to_dumps="${homedir}/${server_config_subpath}/dumps"
    fi

    # add systemd serivce unit file
    local service_file="${systemd_path}/${g_service}${service_suffix}.service"
    sed "s|%username%|${username}|g;s|%path_to_config%|${absolute_path_to_config}|g;s|%path_to_dumps%|${absolute_path_to_dumps}|g;" ${g_service}.service > ${service_file} 

    if [ -n "${service_suffix}" ]; then
        g_service="${g_service}${service_suffix}" # Set just registered non-standart service name for further operations
    fi

    # if service is running then restart service
    systemctl daemon-reload
    ctlcmd status && ctlcmd restart || true

    # add sudoers to control videoserver without sudo rights
    if [ "${g_service}" == "ivideon-server" ]; then
        add_sudoers ${username}
    fi
}

do_enable() {
    ctlcmd enable
}

do_run() {
    ctlcmd start
}

do_stop() {
    ctlcmd stop || true
}

do_disable() {
    ctlcmd disable || true
}

do_uninstall() {
    if check_service ; then
        # remove sudoers file
        if [ "${service}" == "ivideon-server" ]; then
            local username=$(ctlcmd cat | grep "User" | awk -F'=' '{ print $2 }')
            rm -f /etc/sudoers.d/${username}-ivideon
        fi
        # remove systemd unit file
        rm -f $(ctlcmd cat | head -n 1 | awk '{ print $2 }')
        systemctl daemon-reload
    fi
}

main() {
    cd "$(dirname "$0")"

    if [[ $EUID -ne 0 && ${1} != "is-service" && ${1} != "check" ]]; then
        usage "This script must be run as root"
    fi

    if [ $# -lt 1 ]; then
        usage "Too few arguments"
    fi

    local action="$1"
    g_service="${2-}"
    shift 1
    local services="$@" # For multiple-operations
    shift 1

    # Test pre-conditions:
    case "${action}" in
        is-service | check | register | install | enable | install_and_run)
            is_service_supported "${g_service}"
            ;;
        disable | stop_and_disable | uninstall)
            if [ -z "${services-}" ] ; then
                services="${g_services[@]}"
            fi

            for s in ${services[@]} ; do
                is_service_supported "${s}"
            done
            ;;
    esac

    case "${action}" in
        is-service)
            check_service && echo "Service registered"
            ;;
        check)
            check_service && check_autorun
            ;;
        register)
            check_service && echo "Service exist, updating it.." || true
            do_register "$@"
            ;;
        install)
            check_service || do_register "$@"
            do_enable
            ;;
        enable)
            check_service && do_enable
            ;;
        install_and_run)
            check_service || do_register "$@"
            do_enable && do_run
            ;;
        disable)
            for g_service in ${services[@]} ; do
                 do_disable
            done
            ;;
        stop_and_disable)
            for g_service in ${services[@]} ; do
                do_stop && do_disable
            done
            ;;
        uninstall)
            for g_service in ${services[@]} ; do
                do_stop && do_disable && do_uninstall 
            done
            ;;
        *)
            usage "Unknown action"
            ;;
    esac
}

main "$@"
